/*  Wilbur - file indexing and search utility
 *  Copyright (C) 1995-1999  Craig Morris and RedTree Development Inc.
 *        craig@redtree.com - <http://wilbur.redtree.com>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *  or at http://www.fsf.org
*/
#include "stdafx.h"

#ifndef __WILBUR_H__
#include "wilbur.h"
#endif

#ifndef __WILBUDOC_H__
#include "wilbudoc.h"
#endif

#ifndef __FILEMEMORY_H__
#include "FileMemory.h"
#endif

IMPLEMENT_SERIAL( CFileMemory, CObject, 0 );


CFileMemory::CFileMemory()
	:
	m_nInitialIncrement( 0 ), m_nMaxIncrement( 0 ), m_nNextNew( 0 ), m_pBytes( NULL ),
	m_nCurrentSize( 0 ), m_bFileOpen( FALSE ), m_bMapOpen( FALSE ),
	m_bTemporaryFile( FALSE )
{
}

CFileMemory::CFileMemory(  MemSize nInitialIncrement, MemSize nMaxIncrement,
			         CString sFileExtension )
	:
	m_nInitialIncrement( nInitialIncrement ), m_nMaxIncrement( nMaxIncrement ),
	m_sExtension( sFileExtension ), m_nNextNew( 0 ), m_pBytes( NULL ),
	m_nCurrentSize( 0 ), m_bFileOpen( FALSE ), m_bMapOpen( FALSE ),
	m_bTemporaryFile( FALSE )
{
}

CFileMemory::~CFileMemory()
{
	if( m_bMapOpen )
		m_fileMap.Close();

	if( m_bFileOpen )
		m_file.Close();

	if( m_bTemporaryFile )
		CFile::Remove( m_sStoredFileName );
}

static void copyFile( CString sFrom, CString sTo )
{
	CFileStatus inStatus, outStatus;
	bool bStatus = false;

	char* pBuffer = NULL;
	{
		CFile fileTo( sTo, CFile::modeCreate | CFile::modeWrite );

		if( sFrom == "" )
			return;  // Nothing to copy from

		CFile fileFrom( sFrom, CFile::modeRead );

		const int nBufferSize = 0xffff;
		pBuffer = new char[ nBufferSize ];

		int nRead = 0;

		try
		{
			while( (nRead = fileFrom.Read( pBuffer, nBufferSize ) ) )
				fileTo.Write( pBuffer, nRead );

			if( fileFrom.GetStatus( inStatus ) && fileTo.GetStatus( outStatus ) )
			{
				// need to reset destination file times so update corruption checks work
				outStatus.m_mtime = inStatus.m_mtime;
				outStatus.m_ctime = inStatus.m_ctime;
				bStatus = true;
			}
		}
		catch( ... )
		{
			delete pBuffer;
			throw;
		}
	}

	delete pBuffer;

	if( bStatus )
		CFile::SetStatus( sTo, outStatus );

}

void CFileMemory::closeFile()
{
	if( m_bMapOpen )
		m_fileMap.Close();

	if( m_bFileOpen )
		m_file.Close();
}

typedef DWORD ReleaseNumber;

void CFileMemory::Serialize( CArchive& archive )
{
	ReleaseNumber nCurrentRelease = 1;
	ReleaseNumber nRelease;

	CDocument* pDoc = archive.m_pDocument;
	ASSERT( pDoc->IsKindOf( RUNTIME_CLASS( CWilburDoc ) ) );

	if( archive.IsStoring() )
	{
		CString sFileName = makeFileName( (CWilburDoc*) pDoc );

		if( sFileName != m_sStoredFileName )
		{
			if( m_bMapOpen )
			{
				m_fileMap.Close();
				m_bMapOpen = FALSE;
			}

			if( m_bFileOpen )
			{
				m_file.Close();
				m_bFileOpen = FALSE;
			}

			if( m_bTemporaryFile )
			{
				try{ CFile::Remove( sFileName ); } catch( ... ){};

				if( !MoveFile( m_sStoredFileName, sFileName ) )
				{
					copyFile( m_sStoredFileName, sFileName );  // shouldn't happen
				}

				m_bTemporaryFile = FALSE;
			}
			else
			{
				copyFile( m_sStoredFileName, sFileName );
			}

			openMapFile( (CWilburDoc*) pDoc );
		}
		else if( m_bFileOpen && m_bMapOpen )
			m_fileMap.Flush();  // seems like reasonable thing to do


		archive << nCurrentRelease;
		archive << m_sExtension << m_nInitialIncrement << m_nInitialIncrement;
		archive << m_nCurrentSize << m_nNextNew;
	}
	else
	{
		archive >> nRelease;
		if( nRelease < 1 && nRelease > nCurrentRelease )
			AfxThrowArchiveException( CArchiveException::badSchema );

		archive >> m_sExtension >> m_nInitialIncrement >> m_nInitialIncrement;
		archive >> m_nCurrentSize >> m_nNextNew;

		m_nMaxIncrement = m_nInitialIncrement;

		m_sExtension = m_sExtension.Left(3); // temporary thing to read old development files
		openMapFile(  (CWilburDoc*) pDoc );
	}
}

void CFileMemory::openMapFile( CWilburDoc* pDoc )
{
	if( m_nCurrentSize )
	{
		CString sFileName = makeFileName( pDoc );

		if( !m_file.Open( sFileName, CFile::modeRead | CFile::shareDenyWrite) )
		{
			CString sMessage;
			sMessage.LoadString( IDS_MEM_FILE_OPEN_ERR );
			sMessage += sFileName;
			simpleMessageBox( sMessage );
			clearAll();
			AfxThrowUserException();
		}
		m_bFileOpen = TRUE;
		m_sStoredFileName = sFileName;

		if( !m_fileMap.Open( NULL, (HANDLE) m_file.m_hFile, PAGE_READONLY, 0,0, m_nCurrentSize, FILE_MAP_READ ) )
		{
			ASSERT( 0 );
			clearAll();
			AfxThrowUserException();
		}
		m_bMapOpen = TRUE;
		
		if( !m_fileMap.MapViewOfFile( FILE_MAP_READ ) )
		{
			ASSERT( 0 );
			m_fileMap.Close();
			clearAll();
			AfxThrowUserException();
		}

		m_pBytes = m_fileMap;
	}
}

BOOL CFileMemory::enableWrite()
{
	if( !m_bMapOpen || !m_bFileOpen )
		return FALSE;

	m_fileMap.Close();
	m_bMapOpen = FALSE;

	m_file.Close();
	m_bFileOpen = FALSE;

	if( !m_file.Open( m_sStoredFileName, CFile::modeReadWrite | CFile::shareExclusive ) )
	{
		disableWrite();
		return FALSE;
	}

	m_bFileOpen = TRUE;

	if( !m_fileMap.Open( NULL, (HANDLE) m_file.m_hFile, PAGE_READWRITE, 0,0,0 ) )
	{
		m_file.Close();
		m_bFileOpen = FALSE;
		return FALSE;
	}

	m_bMapOpen = TRUE;

	if( !m_fileMap.MapViewOfFile( FILE_MAP_WRITE ) )
	{
		ASSERT( 0 );
		m_fileMap.Close();
		m_bMapOpen = FALSE;
		clearAll();
		AfxThrowUserException();
	}

	m_pBytes = m_fileMap;

	return TRUE;
}

BOOL CFileMemory::disableWrite()
{
	if( m_bMapOpen )
	{
		m_fileMap.Close();
		m_bMapOpen = FALSE;
	}

	if( m_bFileOpen )
	{
		m_file.Close();
		m_bFileOpen = FALSE;
	}

	if( !m_file.Open( m_sStoredFileName,  CFile::modeRead | CFile::shareDenyWrite) )
		return FALSE;
	m_bFileOpen = TRUE;

	if( !m_fileMap.Open( NULL, (HANDLE) m_file.m_hFile, PAGE_READONLY, 0,0, m_nCurrentSize, FILE_MAP_READ ) )
	{
		m_file.Close();
		m_bFileOpen = FALSE;
		return FALSE;
	}

	m_bMapOpen = TRUE;

	if( !m_fileMap.MapViewOfFile( FILE_MAP_READ ) )
	{
		ASSERT( 0 );
		m_fileMap.Close();
		m_bMapOpen = FALSE;
		clearAll();
		AfxThrowUserException();
	}

	m_pBytes = m_fileMap;
	return TRUE;
}

CFile* CFileMemory::createTemporaryFile()
{
	ASSERT( !m_bFileOpen && !m_bMapOpen );
	char pszTempFileName[ MAX_PATH ];;

	if( !GetTempFileName( thisApp()->getZipPath(), m_sExtension, 0, pszTempFileName ) )
	{
		DWORD nErr = GetLastError();

		char* pszMessage;
		FormatMessage( FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM,
					   NULL, nErr, 0, (char*) &pszMessage, 10, NULL );
		simpleMessageBox( pszMessage );

		AfxThrowUserException();
	}

	if( !m_file.Open( pszTempFileName, CFile::modeReadWrite | CFile::shareDenyNone ) )
	{
		CString sMessage;
		sMessage.LoadString( IDS_MEM_FILE_OPEN_ERR );
		sMessage += pszTempFileName;
		simpleMessageBox( sMessage );
		AfxThrowUserException();
	}

	m_sStoredFileName = pszTempFileName;
	m_bTemporaryFile = TRUE;
	m_bFileOpen = TRUE;
	m_nNextNew = 0;
	m_pBytes = NULL;

	return &m_file;
}

void CFileMemory::mapFile( MemSize nSize )
{
	if( nSize )
	{
		m_nCurrentSize = nSize;
		m_nNextNew = nSize + 1;
	}

	if( !m_bFileOpen )
	{
		if( !m_file.Open( m_sStoredFileName, CFile::modeRead | CFile::shareDenyWrite) )
		{
			CString sMessage;
			sMessage.LoadString( IDS_MEM_FILE_OPEN_ERR );
			sMessage += m_sStoredFileName;
			simpleMessageBox( sMessage );
			clearAll();
			AfxThrowUserException();
		}
		m_bFileOpen = TRUE;
	}

	if( !m_fileMap.Open( NULL, (HANDLE) m_file.m_hFile, PAGE_READONLY, 0,0, m_nCurrentSize, FILE_MAP_READ ) )
	{
		ASSERT( 0 );
		clearAll();
		AfxThrowUserException();
	}
	m_bMapOpen = TRUE;
	
	if( !m_fileMap.MapViewOfFile( FILE_MAP_READ ) )
	{
		ASSERT( 0 );
		m_fileMap.Close();
		clearAll();
		AfxThrowUserException();
	}

	m_pBytes = m_fileMap;
}

MemOffset CFileMemory::getNewSpace( MemSize nBytes )
{
	if( nBytes > m_nMaxIncrement )
	{
		// sanity check - if this happens, it is probably not good news
		ASSERT( 0 );
		CString sMsg;
		sMsg.Format( IDS_GET_HUGE_SPACE, nBytes );
		simpleMessageBox( sMsg );
		AfxThrowUserException();
	}

	if( m_nNextNew + nBytes > m_nCurrentSize )
	{
		MemSize nOldSize = m_nCurrentSize;
		if( m_pBytes )
		{
			m_fileMap.Close();
			m_bMapOpen = FALSE;
			m_pBytes = NULL;
		}

		if( !m_bFileOpen )
		{
			char pszTempFileName[ MAX_PATH ];;

			if( !GetTempFileName( thisApp()->getZipPath(), m_sExtension, 0, pszTempFileName ) )
			{
				DWORD nErr = GetLastError();

				char* pszMessage;
				FormatMessage( FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM,
							   NULL, nErr, 0, (char*) &pszMessage, 10, NULL );
				simpleMessageBox( pszMessage );

				AfxThrowUserException();
			}

			if( !m_file.Open( pszTempFileName, CFile::modeReadWrite | CFile::shareDenyNone ) )
			{
				CString sMessage;
				sMessage.LoadString( IDS_MEM_FILE_OPEN_ERR );
				sMessage += pszTempFileName;
				simpleMessageBox( sMessage );
				AfxThrowUserException();
			}

			m_sStoredFileName = pszTempFileName;
			m_bTemporaryFile = TRUE;
			m_bFileOpen = TRUE;
		}


		if( m_nCurrentSize )
		{
			if( m_nCurrentSize < m_nMaxIncrement )
				m_nCurrentSize *= 2;
			else
				m_nCurrentSize += m_nMaxIncrement;
		}
		else
			m_nCurrentSize = m_nInitialIncrement;

		if( m_nNextNew + nBytes > m_nCurrentSize )
			m_nCurrentSize = m_nNextNew + nBytes;

		try
		{
			m_file.SetLength( m_nCurrentSize );
		}
		catch( ... )
		{
			clearAll();
			simpleMessageBox( IDS_MEM_FILE_EXTEND_ERR );
			AfxThrowUserException();  // don't allow generic throw, want to identify exception requiring an abort
		}

		if( !m_fileMap.Open( NULL, (HANDLE) m_file.m_hFile, PAGE_READWRITE, 0,0,0 ) )
		{
			ASSERT( 0 );
			clearAll();
			AfxThrowUserException();
		}

		m_bMapOpen = TRUE;
		
		if( !m_fileMap.MapViewOfFile( FILE_MAP_WRITE ) )
		{
			ASSERT( 0 );
			m_fileMap.Close();
			m_bMapOpen = FALSE;
			clearAll();
			AfxThrowUserException();
		}

		m_pBytes = m_fileMap;
		memset( m_pBytes + nOldSize, 0, m_nCurrentSize - nOldSize );
	}

	MemOffset nReturn = m_nNextNew;
	m_nNextNew += nBytes;

	return nReturn;
}

BYTE* CFileMemory::getPointer( MemOffset nOffset )
{
	if( m_pBytes && nOffset < m_nNextNew )
		return m_pBytes + nOffset;
	else
		return NULL;
}

void CFileMemory::clearAll()
{
	m_nCurrentSize = 0;
	m_nNextNew = 0;
	m_pBytes = NULL;

	if( m_bMapOpen )
	{
		m_fileMap.Close();
		m_bMapOpen = FALSE;
	}

	if( m_bFileOpen )
	{
		m_file.Close();
		m_bFileOpen = FALSE;
	}
}

CString CFileMemory::makeFileName( CWilburDoc* pDoc )
{
	CString sPath = pDoc->getDocFileName();
	int nDot = sPath.ReverseFind( '.' );
	if( nDot == -1 )
	{
		sPath += '.';
		return sPath + m_sExtension;
	}

	return sPath.Left( nDot + 1) + m_sExtension;
}

void CFileMemory::reportError(const char* pszWhere )
{
	CString sReport;
	sReport.Format( IDS_MAPPED_MEM_REQ_ERR,
		pszWhere, m_bMapOpen, m_sExtension, m_nCurrentSize,m_nNextNew,m_pBytes);

	simpleMessageBox( sReport );

	
	AfxThrowMemoryException();
}